@@ -1,5 +1,5 @@
 /* This file is part of Metaproxy.
-   Copyright (C) 2005-2010 Index Data
+   Copyright (C) 2005-2012 Index Data
 
 Metaproxy is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free
@@ -73,7 +73,8 @@ static void die(std::string s) { throw mp::filter::FilterException(s); }
 
 
 // Read XML config.. Put config info in m_p.
-void mp::filter::AuthSimple::configure(const xmlNode * ptr, bool test_only)
+void mp::filter::AuthSimple::configure(const xmlNode * ptr, bool test_only,
+                                       const char *path)
 {
     std::string userRegisterName;
     std::string targetRegisterName;
