@@ -1,5 +1,5 @@
 /* This file is part of Metaproxy.
-   Copyright (C) 2005-2010 Index Data
+   Copyright (C) 2005-2012 Index Data
 
 Metaproxy is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free
@@ -18,6 +18,8 @@ Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 #include "config.hpp"
 
+#include <sstream>
+#include <iomanip>
 #include <metaproxy/util.hpp>
 #include "pipe.hpp"
 #include <metaproxy/filter.hpp>
@@ -28,7 +30,9 @@ Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 #include <yazpp/pdu-assoc.h>
 #include <yazpp/socket-manager.h>
 #include <yazpp/limit-connect.h>
+#include <yaz/timing.h>
 #include <yaz/log.h>
+#include "gduutil.hpp"
 
 #include <iostream>
 
@@ -38,13 +42,20 @@ namespace metaproxy_1 {
     class My_Timer_Thread;
     class ZAssocServer;
     namespace filter {
+        class FrontendNet::Port {
+            friend class Rep;
+            friend class FrontendNet;
+            std::string port;
+            std::string route;
+        };
         class FrontendNet::Rep {
             friend class FrontendNet;
             int m_no_threads;
-            std::vector<std::string> m_ports;
+            std::vector<Port> m_ports;
             int m_listen_duration;
             int m_session_timeout;
             int m_connect_max;
+            std::string m_msg_config;
             yazpp_1::SocketManager mySocketManager;
             ZAssocServer **az;
         };
@@ -63,9 +74,12 @@ namespace metaproxy_1 {
     public:
         ~ZAssocChild();
         ZAssocChild(yazpp_1::IPDU_Observable *the_PDU_Observable,
-                          mp::ThreadPoolSocketObserver *m_thread_pool_observer,
-                          const mp::Package *package);
+                    mp::ThreadPoolSocketObserver *m_thread_pool_observer,
+                    const mp::Package *package,
+                    std::string route,
+                    const char *msg_config);
         int m_no_requests;
+        std::string m_route;
     private:
         yazpp_1::IPDU_Observer* sessionNotify(
             yazpp_1::IPDU_Observable *the_PDU_Observable,
@@ -81,25 +95,28 @@ namespace metaproxy_1 {
         mp::Origin m_origin;
         bool m_delete_flag;
         const mp::Package *m_package;
+        const char *m_msg_config;
     };
     class ThreadPoolPackage : public mp::IThreadPoolMsg {
     public:
-        ThreadPoolPackage(mp::Package *package, mp::ZAssocChild *ses) :
-            m_assoc_child(ses), m_package(package) { };
+        ThreadPoolPackage(mp::Package *package, mp::ZAssocChild *ses,
+            const char *msg_config);
         ~ThreadPoolPackage();
         IThreadPoolMsg *handle();
-        void result();
-        
+        void result(const char *t_info);
+        bool cleanup(void *info);
     private:
+        yaz_timing_t timer;
         mp::ZAssocChild *m_assoc_child;
         mp::Package *m_package;
-        
+        const char *m_msg_config;        
     };
     class ZAssocServer : public yazpp_1::Z_Assoc {
     public:
         ~ZAssocServer();
         ZAssocServer(yazpp_1::IPDU_Observable *PDU_Observable, int timeout,
-                     int connect_max);
+                     int connect_max, std::string route,
+                     const char *msg_config);
         void set_package(const mp::Package *package);
         void set_thread_pool(ThreadPoolSocketObserver *m_thread_pool_observer);
     private:
@@ -110,22 +127,43 @@ namespace metaproxy_1 {
         
         void failNotify();
         void timeoutNotify();
-    void connectNotify();
+        void connectNotify();
     private:
         mp::ThreadPoolSocketObserver *m_thread_pool_observer;
         const mp::Package *m_package;
         int m_session_timeout;
         int m_connect_max;
         yazpp_1::LimitConnect limit_connect;
+        std::string m_route;
+        const char *m_msg_config;
     };
 }
 
+mp::ThreadPoolPackage::ThreadPoolPackage(mp::Package *package,
+                                         mp::ZAssocChild *ses,
+                                         const char *msg_config) :
+    m_assoc_child(ses), m_package(package), m_msg_config(msg_config)
+{
+    if (msg_config)
+        timer = yaz_timing_create();
+    else
+        timer = 0;
+}
+
 mp::ThreadPoolPackage::~ThreadPoolPackage()
 {
+    yaz_timing_destroy(&timer); // timer may be NULL
     delete m_package;
 }
 
-void mp::ThreadPoolPackage::result()
+bool mp::ThreadPoolPackage::cleanup(void *info)
+{
+    mp::Session *ses = (mp::Session *) info;
+
+    return *ses == m_package->session();
+}
+
+void mp::ThreadPoolPackage::result(const char *t_info)
 {
     m_assoc_child->m_no_requests--;
 
@@ -170,25 +208,48 @@ void mp::ThreadPoolPackage::result()
     {
         m_assoc_child->close();
     }
+
+    if (m_msg_config)
+    {
+        yaz_timing_stop(timer);
+        double duration = yaz_timing_get_real(timer);
+        Z_GDU *z_gdu = gdu->get();
+
+        std::ostringstream os;
+        os  << m_msg_config << " "
+            << *m_package << " "
+            << std::fixed << std::setprecision (6) << duration << " ";
+
+        if (z_gdu) 
+            os << *z_gdu;
+        else
+            os << "-";
+        
+        yaz_log(YLOG_LOG, "%s %s", os.str().c_str(), t_info);
+    }
+
     delete this;
 }
 
 mp::IThreadPoolMsg *mp::ThreadPoolPackage::handle() 
 {
-    m_package->move();
+    m_package->move(m_assoc_child->m_route);
     return this;
 }
 
 
 mp::ZAssocChild::ZAssocChild(yazpp_1::IPDU_Observable *PDU_Observable,
-				     mp::ThreadPoolSocketObserver *my_thread_pool,
-				     const mp::Package *package)
-    :  Z_Assoc(PDU_Observable)
+                             mp::ThreadPoolSocketObserver *my_thread_pool,
+                             const mp::Package *package,
+                             std::string route,
+                             const char *msg_config)
+    :  Z_Assoc(PDU_Observable), m_msg_config(msg_config)
 {
     m_thread_pool_observer = my_thread_pool;
     m_no_requests = 0;
     m_delete_flag = false;
     m_package = package;
+    m_route = route;
     const char *peername = PDU_Observable->getpeername();
     if (!peername)
         peername = "unknown";
@@ -212,9 +273,23 @@ void mp::ZAssocChild::recv_GDU(Z_GDU *z_pdu, int len)
 
     mp::Package *p = new mp::Package(m_session, m_origin);
 
-    mp::ThreadPoolPackage *tp = new mp::ThreadPoolPackage(p, this);
-    p->copy_filter(*m_package);
+    mp::ThreadPoolPackage *tp = new mp::ThreadPoolPackage(p, this,
+                                                          m_msg_config);
+    p->copy_route(*m_package);
     p->request() = yazpp_1::GDU(z_pdu);
+
+    if (m_msg_config)
+    {
+        if (z_pdu)          
+        {
+            std::ostringstream os;
+            os  << m_msg_config << " "
+                << *p << " "
+                << "0.000000" << " " 
+                << *z_pdu;
+            yaz_log(YLOG_LOG, "%s", os.str().c_str());
+        }
+    }
     m_thread_pool_observer->put(tp);  
 }
 
@@ -233,9 +308,11 @@ void mp::ZAssocChild::failNotify()
 
     mp::Package *p = new mp::Package(m_session, m_origin);
 
-    mp::ThreadPoolPackage *tp = new mp::ThreadPoolPackage(p, this);
-    p->copy_filter(*m_package);
-    m_thread_pool_observer->put(tp);  
+    mp::ThreadPoolPackage *tp = new mp::ThreadPoolPackage(p, this,
+                                                          m_msg_config);
+    p->copy_route(*m_package);
+    m_thread_pool_observer->cleanup(tp, &m_session);
+    m_thread_pool_observer->put(tp);
 }
 
 void mp::ZAssocChild::timeoutNotify()
@@ -249,9 +326,10 @@ void mp::ZAssocChild::connectNotify()
 }
 
 mp::ZAssocServer::ZAssocServer(yazpp_1::IPDU_Observable *PDU_Observable,
-                               int timeout, int connect_max)
+                               int timeout, int connect_max,
+                               std::string route, const char *msg_config)
     :  Z_Assoc(PDU_Observable), m_session_timeout(timeout),
-       m_connect_max(connect_max)
+       m_connect_max(connect_max), m_route(route), m_msg_config(msg_config)
 {
     m_package = 0;
 }
@@ -282,7 +360,7 @@ yazpp_1::IPDU_Observer *mp::ZAssocServer::sessionNotify(yazpp_1::IPDU_Observable
     }
     mp::ZAssocChild *my =
 	new mp::ZAssocChild(the_PDU_Observable, m_thread_pool_observer,
-                             m_package);
+                            m_package, m_route, m_msg_config);
     my->timeout(m_session_timeout);
     return my;
 }
@@ -312,6 +390,7 @@ mp::filter::FrontendNet::FrontendNet() : m_p(new Rep)
     m_p->m_no_threads = 5;
     m_p->m_listen_duration = 0;
     m_p->m_session_timeout = 300; // 5 minutes
+    m_p->m_connect_max = 0;
     m_p->az = 0;
 }
 
@@ -324,6 +403,17 @@ mp::filter::FrontendNet::~FrontendNet()
             delete m_p->az[i];
         delete [] m_p->az;
     }
+    m_p->az = 0;
+}
+
+void mp::filter::FrontendNet::stop() const
+{
+    if (m_p->az)
+    {
+        size_t i;
+        for (i = 0; i<m_p->m_ports.size(); i++)
+            m_p->az[i]->server("");
+    }
 }
 
 bool mp::My_Timer_Thread::timeout()
@@ -366,26 +456,37 @@ void mp::filter::FrontendNet::process(Package &package) const
     }
     while (m_p->mySocketManager.processEvent() > 0)
     {
+        int no = m_p->mySocketManager.getNumberOfObservers();
+        if (no <= 1)
+            break;
 	if (tt && tt->timeout())
 	    break;
     }
     delete tt;
 }
 
-void mp::filter::FrontendNet::configure(const xmlNode * ptr, bool test_only)
+void mp::filter::FrontendNet::configure(const xmlNode * ptr, bool test_only,
+                                        const char *path)
 {
     if (!ptr || !ptr->children)
     {
         throw mp::filter::FilterException("No ports for Frontend");
     }
-    std::vector<std::string> ports;
+    std::vector<Port> ports;
     for (ptr = ptr->children; ptr; ptr = ptr->next)
     {
         if (ptr->type != XML_ELEMENT_NODE)
             continue;
         if (!strcmp((const char *) ptr->name, "port"))
         {
-            std::string port = mp::xml::get_text(ptr);
+            Port port;
+            const struct _xmlAttr *attr;
+            for (attr = ptr->properties; attr; attr = attr->next)
+            {
+                if (!strcmp((const char *) attr->name, "route"))
+                    port.route = mp::xml::get_text(attr);
+            }
+            port.port = mp::xml::get_text(ptr);
             ports.push_back(port);
             
         }
@@ -409,7 +510,11 @@ void mp::filter::FrontendNet::configure(const xmlNode * ptr, bool test_only)
         }
         else if (!strcmp((const char *) ptr->name, "connect-max"))
         {
-            m_p->m_connect_max = mp::xml::get_int(ptr->children, 0);
+            m_p->m_connect_max = mp::xml::get_int(ptr, 0);
+        }
+        else if (!strcmp((const char *) ptr->name, "message"))
+        {
+            m_p->m_msg_config = mp::xml::get_text(ptr);
         }
         else
         {
@@ -424,6 +529,23 @@ void mp::filter::FrontendNet::configure(const xmlNode * ptr, bool test_only)
 }
 
 void mp::filter::FrontendNet::set_ports(std::vector<std::string> &ports)
+{
+    std::vector<Port> nports;
+    size_t i;
+
+    for (i = 0; i < ports.size(); i++)
+    {
+        Port nport;
+
+        nport.port = ports[i];
+
+        nports.push_back(nport);
+    }
+    set_ports(nports);
+}
+
+
+void mp::filter::FrontendNet::set_ports(std::vector<Port> &ports)
 {
     m_p->m_ports = ports;
     
@@ -439,11 +561,14 @@ void mp::filter::FrontendNet::set_ports(std::vector<std::string> &ports)
         // create ZAssoc with PDU Assoc
         m_p->az[i] = new mp::ZAssocServer(as, 
                                           m_p->m_session_timeout,
-                                          m_p->m_connect_max);
-        if (m_p->az[i]->server(m_p->m_ports[i].c_str()))
+                                          m_p->m_connect_max,
+                                          m_p->m_ports[i].route,
+                                          m_p->m_msg_config.length() > 0 ?
+                                          m_p->m_msg_config.c_str() : 0);
+        if (m_p->az[i]->server(m_p->m_ports[i].port.c_str()))
         {
             throw mp::filter::FilterException("Unable to bind to address " 
-                                              + std::string(m_p->m_ports[i]));
+                                              + std::string(m_p->m_ports[i].port));
         }
     }
 }
