@@ -1,5 +1,5 @@
 /* This file is part of Metaproxy.
-   Copyright (C) 2005-2010 Index Data
+   Copyright (C) 2005-2012 Index Data
 
 Metaproxy is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free
@@ -30,14 +30,18 @@ namespace metaproxy_1 {
     namespace filter {
         class FrontendNet : public Base {
             class Rep;
+            class Port;
             boost::scoped_ptr<Rep> m_p;
         public:
             FrontendNet();
             ~FrontendNet();
             void process(metaproxy_1::Package & package) const;
-            void configure(const xmlNode * ptr, bool test_only);
+            void configure(const xmlNode * ptr, bool test_only,
+                           const char *path);
+            void stop() const;
         public:
             /// set ports
+            void set_ports(std::vector<Port> &ports);
             void set_ports(std::vector<std::string> &ports);
             // set liten duraction (number of seconcds to listen)
             void set_listen_duration(int d);
