@@ -1,5 +1,5 @@
 /* This file is part of Metaproxy.
-   Copyright (C) 2005-2010 Index Data
+   Copyright (C) 2005-2012 Index Data
 
 Metaproxy is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free
@@ -31,6 +31,7 @@ Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 #include <yaz/otherinfo.h>
 #include <yaz/diagbib1.h>
 #include <yaz/match_glob.h>
+#include <yaz/log.h>
 
 #include <map>
 #include <iostream>
@@ -51,6 +52,7 @@ namespace metaproxy_1 {
         };
         struct VirtualDB::Map {
             Map(std::string database, std::list<std::string> targets, std::string route);
+            Map(std::string database, std::string target, std::string route);
             Map();
             bool match(const std::string db) const;
             std::string m_dbpattern;
@@ -78,6 +80,7 @@ namespace metaproxy_1 {
             void search(Package &package, Z_APDU *apdu);
             void present(Package &package, Z_APDU *apdu);
             void scan(Package &package, Z_APDU *apdu);
+            int relay_apdu(Package &package, Z_APDU *apdu);
 
             void close(Package &package);
             typedef std::map<std::string,VirtualDB::Set>::iterator Sets_it;
@@ -106,6 +109,7 @@ namespace metaproxy_1 {
             
             FrontendPtr get_frontend(Package &package);
             void release_frontend(Package &package);
+            void refresh_torus();
         private:
             std::list<VirtualDB::Map>m_maps;
             typedef std::map<std::string,VirtualDB::Set>::iterator Sets_it;
@@ -173,7 +177,7 @@ yf::VirtualDB::BackendPtr yf::VirtualDB::Frontend::create_backend_from_databases
         // see if we have a route conflict.
         if (!first_route && b->m_route != map_it->m_route)
         {
-            // we have a conflict.. 
+            // we have a conflict.. routing must be same for all
             error_code =  YAZ_BIB1_COMBI_OF_SPECIFIED_DATABASES_UNSUPP;
             BackendPtr ptr;
             return ptr;
@@ -211,6 +215,8 @@ yf::VirtualDB::BackendPtr yf::VirtualDB::Frontend::init_backend(
     req->implementationId = org_init->implementationId;
     req->implementationName = org_init->implementationName;
     req->implementationVersion = org_init->implementationVersion;
+    *req->preferredMessageSize = *org_init->preferredMessageSize;
+    *req->maximumRecordSize = *org_init->maximumRecordSize;
 
     ODR_MASK_SET(req->options, Z_Options_search);
     ODR_MASK_SET(req->options, Z_Options_present);
@@ -227,21 +233,20 @@ yf::VirtualDB::BackendPtr yf::VirtualDB::Frontend::init_backend(
 
     Z_GDU *gdu = init_package.response().get();
     // we hope to get an init response
-    if (gdu && gdu->which == Z_GDU_Z3950 && gdu->u.z3950->which ==
-        Z_APDU_initResponse)
+    error_code = YAZ_BIB1_DATABASE_UNAVAILABLE;
+    if (gdu && gdu->which == Z_GDU_Z3950
+        && gdu->u.z3950->which == Z_APDU_initResponse)
     {
         Z_InitResponse *res = gdu->u.z3950->u.initResponse;
         if (ODR_MASK_GET(res->options, Z_Options_namedResultSets))
         {
             b->m_named_result_sets = true;
         }
-        if (*res->result)
+        if (*res->result && !init_package.session().is_closed())
         {
             m_backend_list.push_back(b);
             return b;
-
         }
-        error_code = YAZ_BIB1_DATABASE_UNAVAILABLE;
         mp::util::get_init_diagnostics(res, error_code, addinfo);
     }
     if (!init_package.session().is_closed())
@@ -267,7 +272,6 @@ void yf::VirtualDB::Frontend::search(mp::Package &package, Z_APDU *apdu_req)
     for (i = 0; i<req->num_databaseNames; i++)
         databases.push_back(req->databaseNames[i]);
 
-    BackendPtr b; // null for now
     Sets_it sets_it = m_sets.find(req->resultSetName);
     if (sets_it != m_sets.end())
     {
@@ -286,37 +290,22 @@ void yf::VirtualDB::Frontend::search(mp::Package &package, Z_APDU *apdu_req)
             return;
         } 
         sets_it->second.m_backend->m_number_of_sets--;
-
-        // pick up any existing backend with a database match
-        std::list<BackendPtr>::const_iterator map_it;
-        map_it = m_backend_list.begin();
-        for (; map_it != m_backend_list.end(); map_it++)
-        {
-            BackendPtr tmp = *map_it;
-            if (tmp->m_frontend_databases == databases)
-                break;
-        }
-        if (map_it != m_backend_list.end()) 
-            b = *map_it;
     }
-    else
+    // pick up any existing database with named result sets ..
+    // or one which has no result sets.. yet.
+    BackendPtr b; // null for now
+    std::list<BackendPtr>::const_iterator map_it;
+    map_it = m_backend_list.begin();
+    for (; map_it != m_backend_list.end(); map_it++)
     {
-        // new result set.
-
-        // pick up any existing database with named result sets ..
-        // or one which has no result sets.. yet.
-        std::list<BackendPtr>::const_iterator map_it;
-        map_it = m_backend_list.begin();
-        for (; map_it != m_backend_list.end(); map_it++)
+        BackendPtr tmp = *map_it;
+        if (tmp->m_frontend_databases == databases &&
+            (tmp->m_named_result_sets ||
+             tmp->m_number_of_sets == 0))
         {
-            BackendPtr tmp = *map_it;
-            if (tmp->m_frontend_databases == databases &&
-                (tmp->m_named_result_sets ||
-                 tmp->m_number_of_sets == 0))
-                break;
-        }
-        if (map_it != m_backend_list.end()) 
             b = *map_it;
+            break;
+        }
     }
     if (!b)  // no backend yet. Must create a new one
     {
@@ -449,6 +438,7 @@ void yf::VirtualDB::Rep::release_frontend(mp::Package &package)
     }
 }
 
+
 yf::VirtualDB::Set::Set(BackendPtr b, std::string setname)
     :  m_backend(b), m_setname(setname)
 {
@@ -470,6 +460,14 @@ yf::VirtualDB::Map::Map(std::string database,
 {
 }
 
+yf::VirtualDB::Map::Map(std::string database, 
+                        std::string target, std::string route)
+    : m_dbpattern(database), m_route(route) 
+{
+    m_targets.push_back(target);
+}
+
+
 yf::VirtualDB::Map::Map()
 {
 }
@@ -616,6 +614,27 @@ void yf::VirtualDB::Frontend::present(mp::Package &package, Z_APDU *apdu_req)
     delete id;
 }
 
+int yf::VirtualDB::Frontend::relay_apdu(mp::Package &package, Z_APDU *apdu_req)
+{
+    int no = 0;
+    std::list<BackendPtr>::const_iterator map_it;
+    map_it = m_backend_list.begin();
+    for (; map_it != m_backend_list.end(); map_it++)
+    {
+        BackendPtr b = *map_it;
+        
+        Package relay_package(b->m_backend_session, package.origin());
+        relay_package.copy_filter(package);
+
+        relay_package.request() = yazpp_1::GDU(apdu_req);
+
+        relay_package.move(b->m_route);
+        package.response() = relay_package.response();
+        no++;
+    }
+    return no;
+}
+
 void yf::VirtualDB::Frontend::scan(mp::Package &package, Z_APDU *apdu_req)
 {
     Z_ScanRequest *req = apdu_req->u.scanRequest;
@@ -668,6 +687,7 @@ void yf::VirtualDB::Frontend::scan(mp::Package &package, Z_APDU *apdu_req)
                                                 &req->num_databaseNames,
                                                 &req->databaseNames);
     }
+
     scan_package.request() = yazpp_1::GDU(apdu_req);
     
     scan_package.move(b->m_route);
@@ -683,8 +703,8 @@ void yf::VirtualDB::Frontend::scan(mp::Package &package, Z_APDU *apdu_req)
 
 
 void yf::VirtualDB::add_map_db2targets(std::string db, 
-                                     std::list<std::string> targets,
-                                     std::string route)
+                                       std::list<std::string> targets,
+                                       std::string route)
 {
     m_p->m_maps.push_back(
         VirtualDB::Map(mp::util::database_name_normalize(db), targets, route));
@@ -692,13 +712,12 @@ void yf::VirtualDB::add_map_db2targets(std::string db,
 
 
 void yf::VirtualDB::add_map_db2target(std::string db, 
-                                    std::string target,
-                                    std::string route)
-{
-    std::list<std::string> targets;
-    targets.push_back(target);
+                                      std::string target,
+                                      std::string route)
 
-    add_map_db2targets(db, targets, route);
+{
+    m_p->m_maps.push_back(
+        VirtualDB::Map(mp::util::database_name_normalize(db), target, route));
 }
 
 void yf::VirtualDB::process(mp::Package &package) const
@@ -751,6 +770,9 @@ void yf::VirtualDB::process(mp::Package &package) const
                 else
                     break;
             
+            *resp->preferredMessageSize = *req->preferredMessageSize;
+            *resp->maximumRecordSize = *req->maximumRecordSize;
+
             package.response() = apdu;
             f->m_is_virtual = true;
         }
@@ -785,7 +807,15 @@ void yf::VirtualDB::process(mp::Package &package) const
         }
         else if (apdu->which == Z_APDU_close)
         {
-            package.session().close();
+            if (f->relay_apdu(package, apdu) == 0)
+            {
+                mp::odr odr;
+                
+                package.response() = odr.create_close(
+                    apdu, Z_Close_finished, "virt_db");
+                
+                package.session().close();
+            }
         }
         else
         {
@@ -801,8 +831,8 @@ void yf::VirtualDB::process(mp::Package &package) const
     m_p->release_frontend(package);
 }
 
-
-void mp::filter::VirtualDB::configure(const xmlNode * ptr, bool test_only)
+void mp::filter::VirtualDB::configure(const xmlNode * ptr, bool test_only,
+                                      const char *path)
 {
     for (ptr = ptr->children; ptr; ptr = ptr->next)
     {
@@ -810,7 +840,7 @@ void mp::filter::VirtualDB::configure(const xmlNode * ptr, bool test_only)
             continue;
         if (!strcmp((const char *) ptr->name, "pass-vhosts"))
         {
-            m_p->pass_vhosts = mp::xml::get_bool(ptr->children, false);
+            m_p->pass_vhosts = mp::xml::get_bool(ptr, false);
         }
         else if (!strcmp((const char *) ptr->name, "virtual"))
         {
@@ -834,7 +864,10 @@ void mp::filter::VirtualDB::configure(const xmlNode * ptr, bool test_only)
                             );
             }
             std::string route = mp::xml::get_route(ptr);
-            add_map_db2targets(database, targets, route);
+
+            VirtualDB::Map vmap(mp::util::database_name_normalize(database),
+                                targets, route);
+            m_p->m_maps.push_back(vmap);
         }
         else
         {
