@@ -1,5 +1,5 @@
 /* This file is part of Metaproxy.
-   Copyright (C) 2005-2010 Index Data
+   Copyright (C) 2005-2012 Index Data
 
 Metaproxy is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free
@@ -24,6 +24,7 @@ Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 #include <yaz/otherinfo.h>
 #include <yaz/querytowrbuf.h>
 #include <yaz/oid_db.h>
+#include <yaz/srw.h>
 
 #include <iostream>
 
@@ -161,11 +162,53 @@ std::string mp_util::database_name_normalize(const std::string &s)
 
 }
 
+void mp_util::piggyback_sr(Z_SearchRequest *sreq,
+                           Odr_int result_set_size,
+                           Odr_int &number_to_present,
+                           const char **element_set_name)
+{
+    Z_ElementSetNames *esn;
+    const char *smallSetElementSetNames = 0;
+    const char *mediumSetElementSetNames = 0;
+
+    esn = sreq->smallSetElementSetNames;
+    if (esn && esn->which == Z_ElementSetNames_generic)
+        smallSetElementSetNames = esn->u.generic;
+
+    esn = sreq->mediumSetElementSetNames;
+    if (esn && esn->which == Z_ElementSetNames_generic)
+        mediumSetElementSetNames = esn->u.generic;
+
+    piggyback(*sreq->smallSetUpperBound,
+              *sreq->largeSetLowerBound,
+              *sreq->mediumSetPresentNumber,
+              smallSetElementSetNames,
+              mediumSetElementSetNames,
+              result_set_size,
+              number_to_present,
+              element_set_name);
+}
+
 void mp_util::piggyback(int smallSetUpperBound,
-                                  int largeSetLowerBound,
-                                  int mediumSetPresentNumber,
-                                  int result_set_size,
-                                  int &number_to_present)
+                        int largeSetLowerBound,
+                        int mediumSetPresentNumber,
+                        int result_set_size,
+                        int &number_to_present)
+{
+    Odr_int tmp = number_to_present;
+    piggyback(smallSetUpperBound, largeSetLowerBound, mediumSetPresentNumber,
+              0, 0, result_set_size, tmp, 0);
+    number_to_present = tmp;
+}
+
+void mp_util::piggyback(Odr_int smallSetUpperBound,
+                        Odr_int largeSetLowerBound,
+                        Odr_int mediumSetPresentNumber,
+                        const char *smallSetElementSetNames,
+                        const char *mediumSetElementSetNames,
+                        Odr_int result_set_size,
+                        Odr_int &number_to_present,
+                        const char **element_set_name)
 {
     // deal with piggyback
 
@@ -173,11 +216,16 @@ void mp_util::piggyback(int smallSetUpperBound,
     {
         // small set . Return all records in set
         number_to_present = result_set_size;
+        if (element_set_name && smallSetElementSetNames)
+            *element_set_name = smallSetElementSetNames;
+            
     }
     else if (result_set_size > largeSetLowerBound)
     {
         // large set . Return no records
         number_to_present = 0;
+        if (element_set_name)
+            *element_set_name = 0;
     }
     else
     {
@@ -185,10 +233,11 @@ void mp_util::piggyback(int smallSetUpperBound,
         number_to_present = mediumSetPresentNumber;
         if (number_to_present > result_set_size)
             number_to_present = result_set_size;
+        if (element_set_name && mediumSetElementSetNames)
+            *element_set_name = mediumSetElementSetNames;
     }
 }
 
-
 bool mp_util::pqf(ODR odr, Z_APDU *apdu, const std::string &q)
 {
     YAZ_PQF_Parser pqf_parser = yaz_pqf_create();
@@ -213,22 +262,19 @@ std::string mp_util::zQueryToString(Z_Query *query)
 {
     std::string query_str = "";
 
-    if (query && query->which == Z_Query_type_1){
+    if (query && query->which == Z_Query_type_1)
+    {
         Z_RPNQuery *rpn = query->u.type_1;
         
-        if (rpn){
-            
-            // allocate wrbuf (strings in YAZ!)
-            WRBUF w = wrbuf_alloc();
+        if (rpn)
+        {
+            mp::wrbuf w;
             
             // put query in w
             yaz_rpnquery_to_wrbuf(w, rpn);
             
             // from w to std::string
-            query_str = std::string(wrbuf_buf(w), wrbuf_len(w));
-            
-            // destroy wrbuf
-            wrbuf_destroy(w);
+            query_str = std::string(w.buf(), w.len());
         }
     }
 
@@ -636,6 +682,50 @@ Z_ReferenceId **mp_util::get_referenceId(const Z_APDU *apdu)
     return 0;
 }
 
+std::string mp_util::uri_encode(std::string s)
+{
+    char *x = (char *) xmalloc(1 + s.length() * 3);
+    yaz_encode_uri_component(x, s.c_str());
+    std::string result(x);
+    xfree(x);
+    return result;
+}
+
+
+std::string mp_util::uri_decode(std::string s)
+{
+    char *x = (char *) xmalloc(1 + s.length());
+    yaz_decode_uri_component(x, s.c_str(), s.length());
+    std::string result(x);
+    xfree(x);
+    return result;
+}
+
+mp::wrbuf::wrbuf()
+{
+    m_wrbuf = wrbuf_alloc();
+}
+
+mp::wrbuf::~wrbuf()
+{
+    wrbuf_destroy(m_wrbuf);
+}
+
+mp::wrbuf::operator WRBUF() const
+{
+    return m_wrbuf;
+}
+
+size_t mp::wrbuf::len()
+{
+    return wrbuf_len(m_wrbuf);
+}
+
+const char *mp::wrbuf::buf()
+{
+    return wrbuf_buf(m_wrbuf);
+}
+
 
 /*
  * Local variables:
